/* This file is generated by glib-genmarshal, do not modify it. This code is licensed under the same license as the containing project. Note that it links to GLib, so must comply with the LGPL linking clauses. */
#pragma once

#include <glib-object.h>

G_BEGIN_DECLS

/* OBJECT:VOID (../xfce4-terminal/terminal/terminal-marshal.list:1) */
G_GNUC_INTERNAL
void _terminal_marshal_OBJECT__VOID (GClosure     *closure,
                                     GValue       *return_value,
                                     guint         n_param_values,
                                     const GValue *param_values,
                                     gpointer      invocation_hint,
                                     gpointer      marshal_data);

/* VOID:OBJECT,INT,INT (../xfce4-terminal/terminal/terminal-marshal.list:2) */
G_GNUC_INTERNAL
void _terminal_marshal_VOID__OBJECT_INT_INT (GClosure     *closure,
                                             GValue       *return_value,
                                             guint         n_param_values,
                                             const GValue *param_values,
                                             gpointer      invocation_hint,
                                             gpointer      marshal_data);


G_END_DECLS
